//
//  FeedListViewController.swift
//  ReviewsFeed
//
//  Created by Dmitrii Ivanov on 21/07/2020.
//  Copyright © 2020 ING. All rights reserved.
//

import Combine
import Foundation
import ReviewsFoundationKit
import ReviewsUIKit
import SwiftUI
import UIKit

final class FeedListViewController: UITableViewController {

    // MARK: Constants
    private let viewModel: ViewModel
    
    // MARK: Properties
    private var cancellables: Set<AnyCancellable> = []
    
    // MARK: Outlets
    private lazy var activityIndicator = {
        let indicator = UIActivityIndicatorView()
        
        indicator.translatesAutoresizingMaskIntoConstraints = false

        return indicator
    }()

    private lazy var filterButton = {
        UIBarButtonItem(
            title: NSLocalizedString(
                .Key.Navigation.Button.filter,
                bundle: .module,
                comment: .empty
            ),
            image: .Icon.filter,
            primaryAction: nil,
            menu: .init(
                title: NSLocalizedString(
                    .Key.Menu.filter,
                    bundle: .module,
                    comment: .empty
                ),
                image: UIImage.Icon.star,
                children: {
                    FilterOption.allCases.map { option -> UIAction in
                        .init(
                            title: option.text,
                            image: .init(systemName: option.icon)
                        ) { [weak self] _ in
                            self?.viewModel.filter(by: option)
                        }
                    }
                }()
            )
        )
    }()
    
    private lazy var pullControl = {
        let control = UIRefreshControl()

        control.attributedTitle = .init(string: NSLocalizedString(
            .Key.PullToRefresh.title,
            bundle: .module,
            comment: .empty
        ))

        control.addTarget(
            self,
            action: #selector(refresh(_:)),
            for: .valueChanged
        )
        
        return control
    }()

    // MARK: Initialisers
    init(_ viewModel: ViewModel) {
        self.viewModel = viewModel

        super.init(style: .plain)
    }
    
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    // MARK: Computed
    var items: [Review] {
        viewModel.items
    }

    // MARK: UIViewController
    override func viewDidLoad() {
        super.viewDidLoad()

        setNavigationBar()
        setView()
        setLayout()
        registerTableCells()
        bindViewModel()
        
        viewModel.fetch()
    }
    
    // MARK: UITableViewDataSource
    override func tableView(
        _ tableView: UITableView,
        numberOfRowsInSection section: Int
    ) -> Int {
        viewModel.itemsCount
    }
    
    override func tableView(
        _ tableView: UITableView,
        cellForRowAt indexPath: IndexPath
    ) -> UITableViewCell {
        if viewModel.isWordsShowing && indexPath.row == 0 {
            makeTopWordsCell(tableView)
        } else {
            makeFeedItemCell(tableView, at: indexPath.row)
        }
    }
    
    // MARK: UITableViewDelegate
    override func tableView(
        _ tableView: UITableView,
        didSelectRowAt indexPath: IndexPath
    ) {
        viewModel.openItem(at: indexPath.row)
        
        tableView.deselectRow(
            at: indexPath,
            animated: true
        )
    }
    
}

// MARK: - Helpers
private extension FeedListViewController {

    // MARK: Actions
    @objc func refresh(_ sender: AnyObject) {
        self.viewModel.fetch()
    }
    
    // MARK: Functions
    func bindViewModel() {
        viewModel.$filter
            .receive(on: RunLoop.main)
            .sink { [weak self] option in
                self?.updateFilterMenu(option)
                self?.tableView.reloadData()
            }
            .store(in: &cancellables)

        viewModel.$isFilterEnabled
            .removeDuplicates()
            .receive(on: RunLoop.main)
            .sink { [weak self] enabled in
                self?.filterButton.isEnabled = enabled
            }
            .store(in: &cancellables)
        
        viewModel.$isLoading
            .filter { $0 == true }
            .receive(on: RunLoop.main)
            .sink { [weak self] _ in
                self?.activityIndicator.startAnimating()
            }
            .store(in: &cancellables)

        viewModel.$isLoading
            .dropFirst()
            .filter { $0 == false }
            .receive(on: RunLoop.main)
            .sink { [weak self] _ in
                self?.activityIndicator.stopAnimating()
                self?.pullControl.endRefreshing()
                self?.tableView.reloadData()
                self?.tableView.scrollToRow(
                    at: .init(row: 0, section: 0),
                    at: .middle,
                    animated: true
                )
            }
            .store(in: &cancellables)
    }
    
    func makeFeedItemCell(
        _ tableView: UITableView,
        at row: Int
    ) -> UITableViewCell {
        guard
            let cell = FeedItemCell.dequeue(from: tableView),
            let item = viewModel.item(for: row)
        else {
            return .init()
        }
        
        cell.separatorInset = .init(width: tableView.bounds.size.width)
        cell.contentConfiguration = {
            if #available(iOS 16.0, *) {
                UIHostingConfiguration {
                    FeedItemView(item)
                }
            } else {
                HostingConfiguration {
                    FeedItemView(item)
                }
            }
        }()
        
        return cell
    }
    
    func makeTopWordsCell(_ tableView: UITableView) -> UITableViewCell {
        guard let cell = TopWordsCell.dequeue(from: tableView) else {
            return .init()
        }
        
        cell.separatorInset = .init(width: tableView.bounds.size.width)
        cell.contentConfiguration = {
            if #available(iOS 16.0, *) {
                UIHostingConfiguration {
                    TopWordsView(viewModel.words)
                }
            } else {
                HostingConfiguration {
                    TopWordsView(viewModel.words)
                }
            }
        }()

        return cell
    }

    func registerTableCells() {
        FeedItemCell.register(in: tableView)
        TopWordsCell.register(in: tableView)
    }
    
    func setLayout() {
        NSLayoutConstraint.activate([
            activityIndicator.centerXAnchor.constraint(equalTo: view.centerXAnchor),
            activityIndicator.centerYAnchor.constraint(equalTo: view.safeAreaLayoutGuide.centerYAnchor),
        ])
    }
    
    func setNavigationBar() {
        navigationController?.navigationBar.prefersLargeTitles = true
        navigationController?.navigationBar.isTranslucent = true

        navigationItem.rightBarButtonItem = filterButton
        navigationItem.title = NSLocalizedString(
            .Key.Navigation.title,
            bundle: .module,
            comment: .empty
        )
    }
    
    func setView() {
        tableView.refreshControl = pullControl
        
        view.addSubview(activityIndicator)
        view.bringSubviewToFront(activityIndicator)
        
        activityIndicator.startAnimating()
    }
    
    func updateFilterMenu(_ option: FilterOption) {
        filterButton
            .menu?
            .children
            .compactMap { $0 as? UIAction }
            .forEach { action in
                action.state = action.title == option.text
                    ? .on
                    : .off
            }
    }
    
}

// MARK: - String+Constants
private extension String {
    enum Cell {
        static let feedItem = "FeedItemCell"
    }
    
    enum Key {
        enum Menu {
            static let filter = "common.filter.menu.title.text"
        }

        enum Navigation {
            static let title = "view.feed-list.navigation-bar.title.text"
            
            enum Button {
                static let filter = "view.feed-list.navigation-bar.button.filter-list.text"
            }
        }
        
        enum PullToRefresh {
            static let title = "view.feed-list.pull-to-refresh.title.text"
        }
    }
}

// MARK: - UIEdgeInsets+Inits
private extension UIEdgeInsets {
    
    // MARK: Initialisers
    init(width: CGFloat) {
        self = .init(
            top: 0,
            left: width,
            bottom: 0,
            right: 0
        )
    }
    
}

// MARK: - Previews
#if DEBUG
import ReviewsiTunesKit

@available(iOS 17.0, *)
#Preview("Feed List loading reviews") {
    MockURLProtocol.response = .init(statusCode: 200)
    
    return UINavigationController(rootViewController: FeedListViewController(.init(
        configuration: .init(session: .mock)
    )))
}

@available(iOS 17.0, *)
#Preview("Feed List with few reviews") {
    MockURLProtocol.response = .init(
        statusCode: 200,
        object: Feed(entries: [
            .init(
                id: 1,
                author: "Some author name #1 here",
                title: "Some review title #1 goes here...",
                content: "Some long, explanatory review comment #1 goes here...",
                rating: 3,
                version: "v1.0.0",
                updated: .init()
            ),
            .init(
                id: 2,
                author: "Some author name #2 here",
                title: "Some review title #2 goes here...",
                content: "Some long, explanatory review comment #2 goes here...",
                rating: 5,
                version: "v1.0.0",
                updated: .init()
            ),
            .init(
                id: 3,
                author: "Some author name #3 here",
                title: "Some review title #3 goes here...",
                content: "Some long, explanatory review comment #3 goes here...",
                rating: 1,
                version: "v1.0.0",
                updated: .init()
            ),
            .init(
                id: 4,
                author: "Some author name #4 here",
                title: "Some review title #4 goes here...",
                content: "Some long, explanatory review comment #4 goes here...",
                rating: 4,
                version: "v1.0.0",
                updated: .init()
            ),
            .init(
                id: 5,
                author: "Some author name #5 here",
                title: "Some review title #5 goes here...",
                content: "Some long, explanatory review comment #5 goes here...",
                rating: 2,
                version: "v1.0.0",
                updated: .init()
            ),
        ])
    )
    
    return UINavigationController(rootViewController: FeedListViewController(.init(
        configuration: .init(session: .mock)
    )))
}

@available(iOS 17.0, *)
#Preview("Feed List with no reviews") {
    MockURLProtocol.response = .init(
        statusCode: 200,
        object: Feed(entries: [])
    )
    
    return UINavigationController(rootViewController: FeedListViewController(.init(
        configuration: .init(session: .mock)
    )))
}

@available(iOS 17.0, *)
#Preview("Feed List with live reviews") {
    return UINavigationController(rootViewController: FeedListViewController(.init()))
}
#endif
